from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError

from ..utils.constants import PDS4_NAMESPACES
from ..utils.logging import logger_init

from ..extern import six

# Safe import of ParseError (Python 2.7 and 3+ only)
try:
    from xml.etree.ElementTree import ParseError

except ImportError:
    ParseError = None

# Initialize the logger
logger = logger_init()

#################################


def read_label(filename, strip_extra_whitespace=True, enforce_default_prefixes=False, include_namespace_map=False):

    """ Reads a PDS4 XML Label into an ``ElementTree`` Element object.

        Parameters
        ----------
        filename : str
            The filename, including the path, of the XML label.
        strip_extra_whitespace : bool, optional
            If True, strips whitespaces (including space, tab and newlines)
            at the beginning and end of element and attribute values, but
            only if value has a single line with non-whitespace characters.
            Defaults to False.
        enforce_default_prefixes : bool, optional
            If True, ensures the default namespace is the pds4 schema,
            ensures that 'disp' is the display dictionary prefix and 'sp'
            is the spectral dictionary prefix. Defaults to False.
        include_namespace_map : bool, optional
            If True, changes method return to a tuple, where the first
            value is the label ElementTree object as usual and the second
            is a `dict` with keys being the namespaces and values being
            the namespace prefixes in this label. Defaults to False.

        Returns
        -------
        ``ElementTree`` Element
            Root element for the read-in PDS4 label

        Notes
        -----
        Python 2 v. Python 3: All strings (tag, text, tail) in returned
        elements are byte strings in Python 2, and unicode strings in
        Python 3.

    """

    # Read-in XML tree
    try:
        xml_tree = ET.iterparse(filename, events=('start-ns', 'end'))
    except IOError:
        raise IOError('Unable to locate or read label file: ' + filename)

    # Adjust XML tree
    try:

        namespace_map = {}

        for event, elem in xml_tree:

            # Add namespace to the namespace map
            if event == 'start-ns':

                if enforce_default_prefixes:

                    # Ensure the PDS4 namespace is the default prefix
                    if elem[1] == PDS4_NAMESPACES['pds4']:
                        elem = ('', elem[1])

                    # Ensure the display dictionary has prefix 'disp'
                    if elem[1] == PDS4_NAMESPACES['disp']:
                        elem = ('disp', elem[1])

                    # Ensure the spectral dictionary has prefix 'sp'
                    if elem[1] == PDS4_NAMESPACES['sp']:
                        elem = ('sp', elem[1])

                # Add namespace to map (different prefix's for an existing URI are skipped)
                if enforce_default_prefixes:
                    if elem[1] not in namespace_map:
                        namespace_map[elem[1]] = elem[0]

                continue

            # For Python 2, we convert unicode meta data to byte strings (native ``str``) such that
            # we do not need to preface things like ``str.format`` and string concatinations with
            # u' everywhere.
            if six.PY2:

                if isinstance(elem.tag, unicode):
                    elem.tag = elem.tag.encode('UTF-8')

                if isinstance(elem.text, unicode):
                    elem.text = elem.text.encode('UTF-8')

                if isinstance(elem.tail, unicode):
                    elem.tail = elem.tail.encode('UTF-8')

            # Strip PDS4 namespace tag (a continuation of ensuring default prefix is PDS4 namespace)
            if (enforce_default_prefixes) and (PDS4_NAMESPACES['pds4'] in elem.tag):
                elem.tag = elem.tag.split('{', 1)[0] + elem.tag.split('}', 1)[1]

            # Strip whitespace in elements and attributes if requested
            if strip_extra_whitespace:

                subiter = ET.ElementTree(elem).getiterator()
                attribs = six.iteritems(elem.attrib)

                # Strip whitespaces at beginning and end of value in elements that do not have children
                if len(elem) == 0:
                    for elem_content in subiter:

                        if (elem_content.text) and (_non_blank_line_count(elem_content.text) == 1):
                            elem_content.text = elem_content.text.strip()

                # Strip whitespaces at beginning and end of attribute values
                for name, value in attribs:
                    if _non_blank_line_count(value) == 1:
                        elem.attrib[name] = value.strip()

        label_xml_root = xml_tree.root

        if include_namespace_map:
            return label_xml_root, namespace_map

    # Raise exception if XML cannot be parsed. In Python 3 we raise from None to avoid confusing re-raise
    except (ExpatError, ParseError):
        six.raise_from(
            ExpatError('The requested PDS4 label file does not appear contain valid XML: ' + filename), None)

    return label_xml_root


def _non_blank_line_count(string):
    """
    Parameters
    ----------
    string : str
        String (potentially multi-line) to search in.

    Returns
    -------
    int
        Number of non-blank lines in string.
    """

    non_blank_counter = 0

    for line in string.splitlines():

        if line.strip():
            non_blank_counter += 1

    return non_blank_counter