from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import math
import platform
import functools
import traceback

from .core import Window, ScrollTextWindow

from ..reader.core import pds4_read
from ..reader.read_tables import table_data_size_check

from ..utils.logging import logger_init

from ..extern import six
from ..extern.six.moves.tkinter import (Menu, Canvas, Frame, Label, Button, BooleanVar)
from ..extern.six.moves.tkinter_tkfiledialog import askopenfilename

# Safe import of numpy
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


class StructureListWindow(Window):
    """ Window that summarizes the structures showing some of their properties and giving buttons to open them """

    def __init__(self, viewer, quiet=False, lazy_load=False):

        # Set initial necessary variables and do other required initialization procedures
        super(StructureListWindow, self).__init__(viewer, withdrawn=False)

        # Set window to not be resizable
        self._widget.resizable(0, 0)

        # Create menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)
        self._add_menu(quiet, lazy_load)

        # Initialize structure list window variables
        self._canvas = None
        self._structure_list = None
        self._label_open = False

    # Opens the label, reads in any structures it contains, calls _draw_summary()
    def open_label(self, filename=None, from_existing_structures=None):

        if (filename is None) and (from_existing_structures is None):
            raise TypeError('Cannot open a label without either a filename or existing StructureList.')

        # Determine whether numpy was imported
        use_numpy = True
        if np is None:
            use_numpy = False

        # Open the label and data, or show an error if one occurs
        try:

            # Lazy-load all structures (to obtain meta-data)
            if from_existing_structures is None:
                self._structure_list = pds4_read(filename, quiet=self.menu_option('quiet'), lazy_load=True,
                                                 use_numpy=use_numpy)

                self._viewer.last_open_dir = os.path.dirname(filename)
                self._set_title("{0} - Data Structure Summary for '{1}'".format(self._widget.title(), filename))

            else:
                self._structure_list = from_existing_structures
                self._set_title("{0} - Data Structure Summary".format(self._widget.title()))

        except Exception as e:

            log = logger.get_handler('log_handler').get_recording()
            ScrollTextWindow(self._viewer, 'An Error Occurred!', log + traceback.format_exc())
            logger.exception(e)

        else:

            # Read data from file (or try to access if it exists) for each structure if lazy-load disabled
            if not self.menu_option('lazy_load'):
                self.reify_structures()

            # Draw the summary window
            self._draw_summary()
            self._label_open = True

    # Read-in the data for each structure
    def reify_structures(self):

        # Initialize logging for errors encountered during data read-in
        exception_occurred = False
        logger.get_handler('log_handler').begin_recording()

        # Get list of unread structures
        unread_structures = [structure for structure in self._structure_list if not structure.data_loaded]

        # Inform user about large structures
        warning_window = None
        large_structures = []
        for structure in unread_structures:

            if structure.is_table() and table_data_size_check(structure, quiet=True):
                large_structures.append(structure)

        if large_structures:
            large_struct_message = 'The following structures: \n\n'

            for structure in large_structures:
                large_struct_message += '{0} structure: {1} \n'.format(structure.type, structure.id)

            large_struct_message += (
                '\ncontain a large amount of data. Loading them may take a while. Recommend lazy-load '
                'be enabled via the Options menu.'
            )

            warning_window = self._issue_warning(large_struct_message, log=False, show=True)

        # Read data for all reasonably small structures, force lazy-load for all others
        for structure in unread_structures:

            try:
                logger.info('Now processing a {0} structure: {1}'.format(structure.type, structure.id))
                structure.data

            except Exception:
                exception_occurred = True
                logger.exception('An error occurred during data read-in.')

        # Close warning window once loading data is finished
        if warning_window:
            warning_window.close()

        # Add logging messages for data read-in to log
        log = logger.get_handler('log_handler').get_recording()
        self._structure_list.read_in_log += '\n' + log

        # Show errors that occurred on loading
        if exception_occurred:
            ScrollTextWindow(self._viewer, 'An Error Occurred!', log)

    # Draws a summary of the opened label onto the screen
    def _draw_summary(self):

        # Add View menu
        view_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='View', menu=view_menu)

        view_menu.add_command(label='Full Label', command=self._open_label)

        view_menu.add_separator()

        view_menu.add_command(label='Read-In Log', command=lambda:
            ScrollTextWindow(self._viewer, 'Label/Data Read-in Log', self._structure_list.read_in_log))

        # Create main canvas (which will contain the header, and a frame for the rest)
        self._canvas = Canvas(self._widget)
        self._canvas.pack(fill='both', expand=1)

        # Create canvas frame and show all structure meta data inside it
        canvas_frame = Frame(self._canvas, takefocus=False, height=31 * len(self._structure_list))
        self._canvas.create_window((0, 0), window=canvas_frame, anchor='nw')
        canvas_frame.grid(row=2, column=0, columnspan=4, sticky='w')

        num_buttons_max = 2

        for i, structure in enumerate(self._structure_list):

            index = Label(canvas_frame, text=i, font=self.get_font())
            canvas_frame.grid_columnconfigure(0, minsize=100)
            index.grid(row=i, column=0, pady=(2, 7))

            name_text = structure.id.strip()
            if len(name_text) >= 30:
                name_text = name_text[:29] + '...'

            name = Label(canvas_frame, text=name_text, font=self.get_font())
            canvas_frame.grid_columnconfigure(1, minsize=200)
            name.grid(row=i, column=1, pady=(2, 7))

            type = Label(canvas_frame, text=structure.type, font=self.get_font())
            canvas_frame.grid_columnconfigure(2, minsize=150)
            type.grid(row=i, column=2, pady=(2, 7))

            button_frame = Frame(canvas_frame)
            button_frame.grid(row=i, column=3, padx=(50, 40), sticky='w')
            font = self.get_font(weight='bold')

            open_label = functools.partial(self._open_label, i)
            view_label = Button(button_frame, text='Label', font=font, width=7, command=open_label)
            view_label.pack(side='left')

            open_table = functools.partial(self._open_table, i)
            view_table = Button(button_frame, text='Table', font=font, width=7, command=open_table)
            view_table.pack(side='left')

            if structure.is_array() and structure.is_displayable():
                open_image = functools.partial(self._open_image, i)
                view_image = Button(button_frame, text='Image', font=font, width=7, command=open_image)
                view_image.pack(side='left')
                num_buttons_max += 1

        # Add the header
        header_frame = Frame(self._canvas, takefocus=False, height=31 * len(self._structure_list))
        header_frame.grid(row=0, column=0, columnspan=4, sticky='w')

        index = Label(header_frame, text='Index', font=self.get_font(10, 'bold'))
        header_frame.grid_columnconfigure(0, minsize=100)
        index.grid(row=0, column=0, pady=(2, 2))

        name = Label(header_frame, text='Name', font=self.get_font(10, 'bold'))
        header_frame.grid_columnconfigure(1, minsize=200)
        name.grid(row=0, column=1, pady=(2, 2))

        type = Label(header_frame, text='Type', font=self.get_font(10, 'bold'))
        header_frame.grid_columnconfigure(2, minsize=150)
        type.grid(row=0, column=2, pady=(2, 2))

        # Add line diving header and summary data
        line_frame = Frame(self._canvas, takefocus=False, height=10)
        line_frame.grid(row=1, column=0, columnspan=4, sticky='w')
        self._widget.update_idletasks()
        self._canvas.create_line(5, 27, self._widget.winfo_width() - 5, 27)

        # Add the View text header now that idletasks has been updated
        view = Label(header_frame, text='View', font=self.get_font(10, 'bold'))
        view_left_pad = math.floor((self._widget.winfo_width() - 100 - 200 - 150) / 2 - 13)
        view_left_pad = view_left_pad if view_left_pad > 0 else 0
        view.grid(row=0, column=3, pady=(2, 0), padx=(view_left_pad, 0))

    # Opens the label view for a structure
    def _open_label(self, structure_idx=None):

        from . import label_view

        if structure_idx is None:
            initial_display = 'full label'
            structure_label = None
        else:
            initial_display = 'object label'
            structure_label = self._structure_list[structure_idx].label

        label_view.open_label(self._viewer, self._structure_list.label, structure_label, initial_display)

    # Opens a table view for a structure
    def _open_table(self, structure_idx):

        from . import table_view

        structure = self._structure_list[structure_idx]

        if 'Table' in structure.type:

            # For large, still unread, tables issue a warning message prior to attempting to open the data
            warning_window = None

            if (not structure.data_loaded) and (table_data_size_check(structure, quiet=True)):
                message = ("{0} structure '{1}' contains a large amount of data. This process may take "
                           "a while.".format(structure.type, structure.id))

                warning_window = self._issue_warning(message, log=False, show=True)

            # Open the table structure
            table_view.open_table(self._viewer, structure)

            # Close warning window once loading is finished
            if warning_window:
                warning_window.close()

        elif 'Array' in structure.type:
            table_view.open_table_from_array(self._viewer, structure)

        else:
            raise TypeError('Cannot show table view of structure having type ' + structure.type)

    # Opens an image view for a structure
    def _open_image(self, structure_idx):

        from . import image_view

        structure = self._structure_list[structure_idx]

        if structure.is_array() and structure.is_displayable():
            image_view.open_image(self._viewer, structure)

        else:
            raise TypeError('Cannot show image view of structure having type ' + structure.type)

    # Dialog window to create a new summary view for another label
    def _open_file_box(self, new_window=True):

        # On Linux, the default filetype selected goes first. On Windows it depends on the version of the
        # askopenfilename dialog being used. There is an old bug in Tkinter, at least under Windows 7, where
        # the older Windows dialog is used; and this older dialog also takes the default filetype first, but
        # the newer dialog takes it last. Ultimately this setting for Windows should be based on the system
        # that the frozen distribution is packaged, such that the correct default filetype is first. On Mac
        # adding any type seems to only allow that type to be selected, so we ignore this option.
        if platform.system() == 'Darwin':
            filetypes = []
        else:
            filetypes = [('XML Files', '.xml'), ('All Files', '*')]

        filename = askopenfilename(title='Open Label',
                                   parent=self._widget,
                                   filetypes=filetypes,
                                   initialdir=self._viewer.last_open_dir)

        # Check that filename is a string type (some OS' return binary str and some unicode for filename).
        # Also check that it is neither None or empty (also depends on OS)
        if isinstance(filename, (six.binary_type, six.text_type)) and (filename.strip() != ''):

            if new_window:
                open_summary(self._viewer, filename=filename,
                             quiet=self.menu_option('quiet'), lazy_load=self.menu_option('lazy_load'))

            else:
                self.reset()
                self.open_label(filename)

    def _add_menu(self, quiet, lazy_load):

        # Initialize menu options
        self._menu_options['quiet'] = BooleanVar()
        self._add_trace(self._menu_options['quiet'], 'w', self._update_quiet, default=quiet)

        self._menu_options['lazy_load'] = BooleanVar()
        self._add_trace(self._menu_options['lazy_load'], 'w', self._update_lazy_load, default=lazy_load)

        # Create file menu
        file_menu = Menu(self._widget, tearoff=0)
        file_menu.add_command(label='Open...', command=lambda: self._open_file_box(False))
        file_menu.add_command(label='Open in New Window...', command=lambda: self._open_file_box(True))

        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Create options menu
        options_menu = Menu(self._widget, tearoff=0)
        options_menu.add_checkbutton(label='Lazy-Load Data', onvalue=True, offvalue=False,
                                     variable=self._menu_options['lazy_load'])
        options_menu.add_checkbutton(label='Hide Warnings', onvalue=True, offvalue=False,
                                     variable=self._menu_options['quiet'])
        self._menu.add_cascade(label='Options', menu=options_menu)

    # Updates the logger state to match current menu options value
    def _update_quiet(self, *args):

        if self.menu_option('quiet'):
            logger.quiet()

        else:
            logger.loud()

    # On disable of lazy-load, loads all data immediately
    def _update_lazy_load(self, *args):

        if self._label_open and (not self.menu_option('lazy_load')):
            self.reify_structures()

    # Resets the window to a state before any data was opened
    def reset(self):

        if self._label_open:

            self._set_title(self._widget.title().split('-')[0])
            self._menu.delete(self._menu.index('last'))
            self._canvas.destroy()
            self._structure_list = None
            self._label_open = False

    def close(self):

        self._structure_list = None
        super(StructureListWindow, self).close()


# Opens a new StructureListWindow, and shows a label summary in it if given a filename or StructureList
def open_summary(viewer, filename=None, from_existing_structures=None, quiet=False, lazy_load=False):

    summary_window = StructureListWindow(viewer, quiet=quiet, lazy_load=lazy_load)

    if (filename is not None) or (from_existing_structures is not None):
        summary_window.open_label(filename=filename, from_existing_structures=from_existing_structures)

    return summary_window
